/**
 * Copyright (C) 2009-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.softlayer.compute;

import com.google.common.base.Function;
import com.google.common.collect.Sets;
import org.apache.log4j.Logger;
import org.dasein.cloud.*;
import org.dasein.cloud.compute.*;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.jclouds.softlayer.SoftLayer;
import org.jclouds.compute.domain.Image;
import org.jclouds.rest.AuthorizationException;
import org.jclouds.rest.RestContext;
import org.jclouds.softlayer.SoftLayerAsyncClient;
import org.jclouds.softlayer.SoftLayerClient;
import org.jclouds.softlayer.compute.functions.ProductItemToImage;
import org.jclouds.softlayer.domain.ProductItem;
import org.jclouds.softlayer.predicates.ProductItemPredicates;

import javax.annotation.Nonnull;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.*;

import static com.google.common.collect.Iterables.filter;
import static com.google.common.collect.Iterables.transform;

public class SoftLayerImage implements MachineImageSupport {

    private SoftLayer provider;

    SoftLayerImage(SoftLayer softlayer) {
        provider = softlayer;
    }

    @Override
    public void remove(String imageId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("GoGrid does not allow you to delete images.");
    }

    @Override
    public Collection<MachineImage> searchMachineImages(String keyword, Platform platform, Architecture architecture) throws InternalException, CloudException {
        //TODO: This code is copied from GoGrid and needs unit tests and refactored.
        ArrayList<MachineImage> result = new ArrayList<MachineImage>();
        for( MachineImage image :  listMachineImages() ) {
            if( architecture != null && !architecture.equals(image.getArchitecture()) ) {
                continue;
            }
            if( platform != null && !platform.equals(Platform.UNKNOWN) ) {
                Platform mine = image.getPlatform();

                if( platform.isWindows() && !mine.isWindows() ) {
                    continue;
                }
                if( platform.isUnix() && !mine.isUnix() ) {
                    continue;
                }
                if( platform.isBsd() && !mine.isBsd() ) {
                    continue;
                }
                if( platform.isLinux() && !mine.isLinux() ) {
                    continue;
                }
                if( platform.equals(Platform.UNIX) ) {
                    if( !mine.isUnix() ) {
                        continue;
                    }
                }
                else if( !platform.equals(mine) ) {
                    continue;
                }
            }
            if( keyword != null && !keyword.equals("") ) {
                keyword = keyword.toLowerCase();
                if( !image.getProviderMachineImageId().toLowerCase().contains(keyword) ) {
                    if( !image.getName().toLowerCase().contains(keyword) ) {
                        if( !image.getDescription().toLowerCase().contains(keyword) ) {
                            continue;
                        }
                    }
                }
            }
            result.add(image);
        }
        return result;
    }


    @Override
    public MachineImage getMachineImage(String imageId) throws InternalException, CloudException {
         for( MachineImage image: listMachineImages() ) {
            if(image.getProviderMachineImageId().equals(imageId)) {
                return image;
            }
        }
        return null;
    }

    /**
     * An 'image' is a product item with a category code 'os'
     */
    @Override
    public String getProviderTermForImage(Locale locale) {
        return "product item";
    }

    @Override
    public Collection<String> listShares(String imageId) throws CloudException, InternalException {
        return new ArrayList<String>();
    }

    @Override
    public AsynchronousTask<String> imageVirtualMachine(String serverId, String name, String description) throws InternalException, CloudException {
        throw new OperationNotSupportedException();
    }

    @Override
    public Collection<MachineImage> listMachineImages() throws InternalException, CloudException {
        Set<ProductItem> items = provider.getProductPackage().getItems();
        if( items == null ) return Collections.emptySet();

        Iterable<MachineImage> images = transform(filter(items, ProductItemPredicates.categoryCode("os")),
                                        toImage());
        return Sets.newLinkedHashSet(images);
    }

    private Function<ProductItem,MachineImage> toImage() {
        return new Function<ProductItem, MachineImage>() {
            public MachineImage apply(ProductItem item) {
                if( item == null ) {
                    return null;
                }
                Image image = new ProductItemToImage().apply(item);

                MachineImage result = new MachineImage();

                result.setCurrentState(MachineImageState.ACTIVE);
                result.setDescription(image.getDescription());
                result.setName(image.getName());
                result.setProviderRegionId(provider.getContext().getRegionId());
                result.setArchitecture(provider.toArchitecture(image.getOperatingSystem()));
                result.setPlatform(provider.toPlatform(image.getOperatingSystem()));
                result.setProviderMachineImageId(image.getId());
                result.setType(MachineImageType.STORAGE);

                result.setProviderOwnerId(null);
                result.setSoftware("");

                if( result.getName() == null ) {
                    result.setName(result.getProviderMachineImageId());
                }
                if( result.getDescription() == null ) {
                    result.setDescription(result.getName());
                }
                return result;
            }
        };
    }

    @Override
    public void downloadImage(String machineImageId, OutputStream toOutput) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public boolean hasPublicLibrary() {
        return true;
    }

    @Override
    public AsynchronousTask<String> imageVirtualMachineToStorage(String vmId, String name, String description, String directory) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public String installImageFromUpload(MachineImageFormat format, InputStream imageStream) throws CloudException, InternalException {
        throw new OperationNotSupportedException("Not supported");
    }

    @Override
    public boolean isImageSharedWithPublic(String machineImageId) throws CloudException, InternalException {
        return false;
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx = provider.getCloudClient();
        try {
            provider.getProductPackage();
            return true;
        }
        catch( AuthorizationException e ) {
            return false;
        }
        catch( RuntimeException e ) {
            return false;
        }
        finally {
            ctx.close();
        }
    }

    /**
     * For softlayer there are no real images - so return all
     */
    @Override
    public Iterable<MachineImage> listMachineImagesOwnedBy(String accountId) throws CloudException, InternalException {
        return listMachineImages();
    }

    @Override
    public Iterable<MachineImageFormat> listSupportedFormats() throws CloudException, InternalException {
        return Collections.emptyList();
    }

    @Override
    public @Nonnull
    String[] mapServiceAction(@Nonnull ServiceAction action) {
        return new String[0];
    }

    @Override
    public String registerMachineImage(String atStorageLocation) throws CloudException, InternalException {
        throw new OperationNotSupportedException("No supported");
    }

    @Override
    public void shareMachineImage(String machineImageId, String withAccountId, boolean allow) throws CloudException, InternalException {
        throw new OperationNotSupportedException("No supported");
    }

    @Override
    public boolean supportsCustomImages() {
        return false;
    }

    @Override
    public boolean supportsImageSharing() {
        return false;
    }

    @Override
    public boolean supportsImageSharingWithPublic() {
        return false;
    }

    @Override
    public String transfer(CloudProvider fromCloud, String machineImageId) throws CloudException, InternalException {
        throw new OperationNotSupportedException("No supported");
    }
    
    static private final Logger counter = Logger.getLogger("gogrid.count");

    private void count(String method) {
        counter.info(method);
    }
}
