/**
 * Copyright (C) 2009-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.softlayer.compute;

import com.google.common.base.Splitter;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Iterables;
import com.google.common.collect.Sets;
import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.Tag;
import org.dasein.cloud.compute.*;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.jclouds.softlayer.SoftLayer;
import org.jclouds.compute.domain.Hardware;
import org.jclouds.compute.domain.Image;
import org.jclouds.predicates.RetryablePredicate;
import org.jclouds.rest.AuthorizationException;
import org.jclouds.rest.RestContext;
import org.jclouds.softlayer.SoftLayerAsyncClient;
import org.jclouds.softlayer.SoftLayerClient;
import org.jclouds.softlayer.compute.functions.ProductItemToImage;
import org.jclouds.softlayer.compute.functions.ProductItems;
import org.jclouds.softlayer.compute.functions.ProductItemsToHardware;
import org.jclouds.softlayer.compute.strategy.SoftLayerComputeServiceAdapter;
import org.jclouds.softlayer.domain.*;
import org.jclouds.softlayer.features.DatacenterClient;
import org.jclouds.softlayer.features.VirtualGuestClient;
import org.jclouds.softlayer.predicates.ProductItemPredicates;

import javax.annotation.Nonnull;
import java.util.*;
import java.util.regex.Pattern;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static com.google.common.base.Predicates.and;
import static com.google.common.collect.Iterables.filter;
import static com.google.common.collect.Iterables.get;
import static org.jclouds.softlayer.predicates.ProductItemPredicates.capacity;
import static org.jclouds.softlayer.predicates.ProductItemPredicates.categoryCode;
import static org.jclouds.softlayer.predicates.ProductItemPredicates.matches;
import static org.jclouds.softlayer.compute.functions.ProductItems.price;

public class SoftLayerServer implements VirtualMachineSupport {
    static private final Logger logger = Logger.getLogger(SoftLayerServer.class);
    static private final String PRODUCT_ID_FORMAT = "%d,%d,%d,%d";
    static private final String PRODUCT_NAME_FORMAT = "cpu:%d,ram:%d,disk:%d,portSpeed:%.0f";
    static private final String PRODUCT_DESCRIPTION_FORMAT = "cpu:%s,ram:%s,disk:%s,portSpeed:%s";

    private final SoftLayer provider;

    private final Pattern cpuPattern;
    private final Pattern disk0Type;
    private final float portSpeed;

    SoftLayerServer(SoftLayer gogrid) {
        provider = gogrid;
        this.cpuPattern = Pattern.compile(checkNotNull(provider.getCpuRegex(), "cpuRegex"));
        this.disk0Type = Pattern.compile(".*" + checkNotNull(provider.getDisk0Type(), "disk0Type") + ".*");
        this.portSpeed = provider.getPortSpeed();
        checkArgument(portSpeed > 0, "portSpeed must be greater than zero, often 10, 100, 1000, 10000");
    }

    /**
     * Boot a server. Must be HALTED or PAUSED. Is a no-op if in another state.
     * @param serverId the id of the server, which is a long for SoftLayer
     * @throws InternalException
     * @throws CloudException
     */
    @Override
    public void boot(String serverId) throws InternalException, CloudException {
        RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx = provider.getCloudClient();
        try {
            VirtualGuestClient client = ctx.getApi().getVirtualGuestClient();
            Long id = Long.parseLong(serverId);
            VirtualGuest guest = client.getVirtualGuest(id);

            switch (guest.getPowerState().getKeyName()) {
                case HALTED:
                    client.powerOnVirtualGuest(id);
                    logger.debug("Powering on halted server:"+serverId);
                    break;
                case PAUSED:
                    client.resumeVirtualGuest(id);
                    logger.debug("Resuming paused server:"+serverId);
                    break;
                default:
                    logger.debug("Server must be HALTED or PAUSED in order to boot. Currently in state:"+guest.getPowerState().getKeyName()+".");
                    break;
            }
        }
        catch( RuntimeException e ) {
            logger.error("Failed to boot server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to boot server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public VirtualMachine clone(String serverId, String intoDcId, String name, String description, boolean powerOn, String ... firewallIds) throws InternalException, CloudException {
        throw new OperationNotSupportedException("SoftLayer does not support server cloning.");
    }

    @Override
    public String getConsoleOutput(String serverId) throws InternalException, CloudException {
        return "";
    }

    @Override
    public Collection<String> listFirewalls(String serverId) throws InternalException, CloudException {
        return Collections.emptyList();
    }

    @Override
    public String getProviderTermForServer(Locale locale) {
        return "virtual guest";
    }

    @Override
    public VirtualMachine getVirtualMachine(String serverId) throws InternalException, CloudException {
        RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx = provider.getCloudClient();
        try {
            return toServer(ctx, getSoftLayerServer(ctx, serverId));
        }
        finally {
            ctx.close();
        }
    }

    private VirtualGuest getSoftLayerServer(RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx, String serverId) throws CloudException {
        try {
            count("getServersById");
            return ctx.getApi().getVirtualGuestClient().getVirtualGuest(Long.parseLong(serverId));
        }
        catch( NumberFormatException nfe ) {
            // Tests supply serverId as strings
            logger.error("Softlayer serverIDs must be long values:" + serverId + ".");
            return null;
        } catch( RuntimeException e ) {
            logger.error("Failed to find server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to find server " + serverId + ": " + e.getMessage());
        }
    }

    @Override
    public VirtualMachineProduct getProduct(String productId) throws InternalException, CloudException {
        for( VirtualMachineProduct product : listProducts(null) ) {
            if( product.getProductId().equals(productId) ) {
                return product;
            }
        }
        return null;
    }
    
    @Override
    public VmStatistics getVMStatistics(String serverId, long begin, long end) throws InternalException, CloudException {
        return new VmStatistics();
    }

    @Override
    public Collection<VmStatistics> getVMStatisticsForPeriod(String serverId, long start, long end) throws InternalException, CloudException {
        return Collections.emptyList();
    }
    
    static private Set<VirtualMachineProduct> cachedProducts;

    /**
     * VirtualMachineProducts are not tied to architecture
     * @param architecture unused
     * @return iterable containing all available VirtualMachineProducts
     * @throws InternalException
     * @throws CloudException
     */
    @Override
    public Iterable<VirtualMachineProduct> listProducts(Architecture architecture) throws InternalException, CloudException {

        if( cachedProducts == null ) {
            cachedProducts = Sets.newLinkedHashSet();
        }

        ProductPackage productPackage = provider.getProductPackage();
        Set<ProductItem> items = productPackage.getItems();

        ProductItem portSpeedItem = Iterables.find(items, and(capacity(portSpeed), categoryCode("port_speed")));
        final int portSpeedPriceId = price().apply(portSpeedItem).getId();

        for (ProductItem cpuItem : filter(items, matches(cpuPattern))) {
           for (ProductItem ramItem : filter(items, categoryCode("ram"))) {
              for (ProductItem sanItem : filter(items, and(matches(disk0Type), categoryCode("guest_disk0")))) {

                  VirtualMachineProduct product = new VirtualMachineProduct();
                  product.setCpuCount(cpuItem.getCapacity().intValue());
                  product.setDiskSizeInGb(sanItem.getCapacity().intValue());
                  product.setRamInMb(ramItem.getCapacity().intValue()*1024);

                  final int cpuPriceId = price().apply(cpuItem).getId();
                  final int ramPriceId = price().apply(ramItem).getId();
                  final int sanPriceId = price().apply(sanItem).getId();
                  String productId = String.format(PRODUCT_ID_FORMAT,cpuPriceId,ramPriceId,sanPriceId,portSpeedPriceId);
                  product.setProductId(productId);

                  product.setName(String.format(PRODUCT_NAME_FORMAT,product.getCpuCount(),product.getRamInMb(),
                                                                    product.getDiskSizeInGb(),portSpeedItem.getCapacity()));
                  product.setDescription(String.format(PRODUCT_DESCRIPTION_FORMAT,cpuItem.getDescription(),ramItem.getDescription(),
                                                                                  sanItem.getDescription(),portSpeedItem.getDescription()));
                  cachedProducts.add(product);
              }
           }
        }
        return cachedProducts;
    }

    @Override
    public VirtualMachine launch(String imageId, VirtualMachineProduct size, String dataCenterId, String serverName, String description, String withKey, String inVlanId, boolean withMonitoring, boolean forImaging, String... firewalls) throws InternalException, CloudException {
        return launch(imageId, size, dataCenterId, serverName, description, withKey, inVlanId, withMonitoring, forImaging, firewalls, new Tag[0]);
    }
    
    public VirtualMachine launch(String imageId, VirtualMachineProduct size, String dataCenterId, String serverName, String description, String withKey, String inVlanId, boolean withMonitoring, boolean forImaging, String[] firewalls, Tag ... tags) throws InternalException, CloudException {
        RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx = provider.getCloudClient();
        try
        {
            VirtualGuestClient client = ctx.getApi().getVirtualGuestClient();
            ProductOrder order = getProductOrder(imageId,size,dataCenterId,serverName);
            logger.debug(String.format("ordering new virtualGuest name(%s)", serverName));
            ProductOrderReceipt productOrderReceipt = client.orderVirtualGuest(order);
            VirtualGuest server = get(productOrderReceipt.getOrderDetails().getVirtualGuests(), 0);
            logger.trace(String.format("virtualGuest(%s)", server.getId()));

            logger.debug(String.format("awaiting login details for virtualGuest(%s)", server.getId()));
            boolean orderInSystem = getLoginTester(ctx.getApi()).apply(server);
            logger.trace(String.format("virtualGuest(%s) complete(%s)", server.getId(), orderInSystem));

            checkState(orderInSystem, String.format("order for guest %s doesn't have login details within %sms", server, Long.toString(provider.getLoginTimeoutMs())));
            server = client.getVirtualGuest(server.getId());

            return toServer(ctx, server);
        }
        catch( RuntimeException e ) {
            throw new CloudException("Failed to define server: " + e.getMessage());
        }
        catch( Error e ) {
            throw new InternalException("Failed to define server: " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    private ProductOrder getProductOrder(String imageId, VirtualMachineProduct product, String dataCenterId, String serverName) throws CloudException {
        VirtualGuest guest = VirtualGuest.builder()
                                         .hostname(serverName)
                                         .domain(provider.getDomain())
                                         .build();

        return ProductOrder.builder()
                           .packageId(provider.getProductPackage().getId())
                           .location(dataCenterId)
                           .quantity(1)
                           .useHourlyPricing(true)
                           .virtualGuest(guest)
                           .prices(getPrices(imageId,product))
                           .build();

    }

    private Iterable<ProductItemPrice> getPrices(String imageId, VirtualMachineProduct product) {
        ImmutableSet.Builder<ProductItemPrice> result = ImmutableSet.<ProductItemPrice> builder();

        // Image ID
        result.add(ProductItemPrice.builder().id(Integer.parseInt(imageId)).build());

        // Configurable IDs
        Iterable<String> hardwareIds = Splitter.on(",").split(product.getProductId());
        for (String hardwareId : hardwareIds) {
            int id = Integer.parseInt(hardwareId);
            result.add(ProductItemPrice.builder().id(id).build());
        }

        // Default IDs
        result.add(ProductItemPrice.builder().id(21).build()); // 1 IP Address
        result.add(ProductItemPrice.builder().id(55).build()); // Host Ping: categoryCode: monitoring, notification
        result.add(ProductItemPrice.builder().id(57).build()); // Email and Ticket: categoryCode: notification
        result.add(ProductItemPrice.builder().id(58).build()); // Automated Notification: categoryCode: response
        result.add(ProductItemPrice.builder().id(1800).build()); // 0 GB Bandwidth: categoryCode: bandwidth
        result.add(ProductItemPrice.builder().id(905).build()); // Reboot / Remote Console: categoryCode: remote_management
        result.add(ProductItemPrice.builder().id(418).build()); // Nessus Vulnerability Assessment & Reporting: categoryCode:vulnerability_scanner
        result.add(ProductItemPrice.builder().id(420).build()); // Unlimited SSL VPN Users & 1 PPTP VPN User per account: categoryCode:vpn_management

        return result.build();
    }

    @Override
    public Collection<VirtualMachine> listVirtualMachines() throws InternalException, CloudException {
        RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx = provider.getCloudClient();

        try {
            Set<VirtualGuest> virtualGuests = ctx.getApi().getVirtualGuestClient().listVirtualGuests();
            Set<VirtualMachine> result = new HashSet<VirtualMachine>(virtualGuests.size());
            for( VirtualGuest guest: virtualGuests ) {
                VirtualMachine virtualMachine = toServer(ctx, guest);
                if( result != null ) {
                    result.add(virtualMachine);
                }
            }
            return result;
        }
        catch( RuntimeException e ) {
            logger.error("Failed to list servers: " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to list servers: " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public boolean supportsAnalytics() throws CloudException, InternalException {
        return false;
    }

    @Override
    public void enableAnalytics(String serverId) throws InternalException, CloudException {
        // NO-OP
    }

    @Override
    public void disableAnalytics(String serverId) throws InternalException, CloudException {
        // NO-OP
    }

    @Override
    public @Nonnull String[] mapServiceAction(@Nonnull ServiceAction action) {
        return new String[0];
    }

    @Override
    public void pause(String serverId) throws InternalException, CloudException {
        RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx = provider.getCloudClient();
        try {
            count("power");
            ctx.getApi().getVirtualGuestClient().pauseVirtualGuest(Long.parseLong(serverId));
        }
        catch( RuntimeException e ) {
            logger.error("Failed to pause server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to pause server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void reboot(String serverId) throws CloudException, InternalException {
        RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx = provider.getCloudClient();

        try {
            count("power");
            ctx.getApi().getVirtualGuestClient().rebootHardVirtualGuest(Long.parseLong(serverId));
        }
        catch( RuntimeException e ) {
            logger.error("Failed to reboot server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to reboot server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void terminate(String serverId) throws InternalException, CloudException {
        RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx = provider.getCloudClient();

        try {
            VirtualGuestClient virtualGuestClient = ctx.getApi().getVirtualGuestClient();
            VirtualGuest guest = virtualGuestClient.getVirtualGuest(Long.parseLong(serverId));
            count("power");

            if(guest!=null && guest.getBillingItemId()>0) {
                virtualGuestClient.cancelService(guest.getBillingItemId());
            }
        }
        catch( RuntimeException e ) {
            logger.error("Failed to delete server " + serverId + ": " + e.getMessage());
            e.printStackTrace();
            throw new CloudException("Failed to delete server " + serverId + ": " + e.getMessage());
        }
        finally {
            ctx.close();
        }
    }

    private VmState toState(VirtualGuest.State serverState) {
        if(serverState ==  null) {
            return VmState.PENDING;
        }
        switch( serverState ) {
            case HALTED: return VmState.PENDING;
            case PAUSED: return VmState.PAUSED;
            case RUNNING: return VmState.RUNNING;
            case UNRECOGNIZED: return VmState.PENDING;
            default: return VmState.RUNNING;
        }
    }

    /**
     * SoftLayer partially populates information when orders are placed and during shutdown.
     */
    private VirtualMachine toServer( RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx, VirtualGuest from) throws InternalException, CloudException {
        if ( from == null ) return null;

        VirtualMachine server = new VirtualMachine();

        final Image image = getImage(ctx,from);
        if ( image!=null ) {
            server.setProviderMachineImageId(String.valueOf(image.getId()));
            server.setArchitecture(provider.toArchitecture(image.getOperatingSystem()));
            server.setPlatform(provider.toPlatform(image.getOperatingSystem()));
        }

        final Region region = getRegion(ctx,from);
        if ( region!=null ) {
            server.setProviderRegionId(region.getKeyname());
        }

        server.setName(from.getFullyQualifiedDomainName());
        server.setProviderVirtualMachineId(String.valueOf(from.getId()));

        final String publicIp = from.getPrimaryIpAddress();
        final String privateIp = from.getPrimaryBackendIpAddress();
        server.setPublicIpAddresses(new String[] { publicIp });
        server.setProviderAssignedIpAddressId(String.valueOf(publicIp) );
        server.setPrivateIpAddresses(new String[] { privateIp });

        server.setCreationTimestamp(from.getCreateDate().getTime());
        server.setCurrentState(toState(from.getPowerState().getKeyName()));
        server.setProviderOwnerId(provider.getContext().getAccountNumber());

        Datacenter datacenter = from.getDatacenter();
        if (datacenter!=null) {
            server.setProviderDataCenterId(String.valueOf(datacenter.getId()));
        }

        server.setDescription("cpu:" + from.getMaxCpu() + ",ram:" + from.getMaxMemory());
        server.setPersistent(true);
        server.setPausable(true);
        server.setRebootable(true);
        server.setTerminationTimestamp(-1L);

        Hardware hardware = getHardware(ctx,from);
        VirtualMachineProduct prd = new VirtualMachineProduct();
        if( hardware != null ) {
            // Cancelled, Hung and server's whose orders are just placed will not have hardware.
            prd.setProductId(hardware.getId());
            prd.setName(hardware.getId());
            prd.setDescription(hardware.getId());
            prd.setCpuCount(hardware.getProcessors().size());
            prd.setRamInMb(hardware.getRam());
            prd.setDiskSizeInGb(getDiskSizeInGb(hardware));
        }
        server.setProduct(prd);

        int billingItemId = from.getBillingItemId();
        if (billingItemId>0) server.addTag("billingItemId",Integer.toString(billingItemId));

        return server;
    }

    private Image getImage(RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx, VirtualGuest guest) {
         // 'bad' orders have no start cpu's and cause the order lookup to fail.
         if (guest.getStartCpus() < 1) {
             return null;
         }
         ProductOrder order = ctx.getApi().getVirtualGuestClient().getOrderTemplate(guest.getId());
         if (order == null)
            return null;
         Iterable<ProductItem> items = Iterables.transform(order.getPrices(), ProductItems.item());
         ProductItem os = Iterables.find(items, ProductItemPredicates.categoryCode("os"));
         return new ProductItemToImage().apply(os);
    }

    private Region getRegion(RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx, VirtualGuest guest ) {
        DatacenterClient datacenterClient = ctx.getApi().getDatacenterClient();
        Datacenter datacenter = datacenterClient.getDatacenter(guest.getDatacenter().getId());
        for( Region region: datacenter.getRegions() ) {
            if ( region.getKeyname().equals("FIRST_AVAILABLE")) continue;
            return region;
        }
        return null;
    }

     private Hardware getHardware(RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx, VirtualGuest guest) {
         // 'bad' orders have no start cpu's and cause the order lookup to fail.
         if (guest.getStartCpus() < 1)
            return null;
         ProductOrder order = ctx.getApi().getVirtualGuestClient().getOrderTemplate(guest.getId());
         if (order == null)
            return null;
         Iterable<ProductItem> items = Iterables.transform(order.getPrices(), ProductItems.item());
         return new ProductItemsToHardware().apply(items);
     }

    private int getDiskSizeInGb(Hardware hardware) {
        int result = 0;
        for( org.jclouds.compute.domain.Volume volume: hardware.getVolumes() ) {
            result += volume.getSize().intValue();
        }
        return result;
    }

    private RetryablePredicate<VirtualGuest> getLoginTester(final SoftLayerClient client) {
        return new RetryablePredicate<VirtualGuest>(
              new SoftLayerComputeServiceAdapter.VirtualGuestHasLoginDetailsPresent(client)
              ,provider.getLoginTimeoutMs());
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        RestContext<SoftLayerClient,SoftLayerAsyncClient> ctx = provider.getCloudClient();
        
        try {
            try {
                ctx.getApi().getVirtualGuestClient().listVirtualGuests();
                return true;
            }
            catch( AuthorizationException e ) {
                e.printStackTrace();
                return false;
            }
            catch( RuntimeException e ) {
                e.printStackTrace();
                logger.error("Failed to determine subscription status: " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException("Failed to determine subscription status: " + e.getMessage());
            }
            catch( Error e ) {
                e.printStackTrace();
                logger.error("Failed to determine subscription status: " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException("Failed to determine subscription status: " + e.getMessage());                
            }
        }
        finally {
            ctx.close();
        }
    }

    static private final Logger counter = Logger.getLogger("softlayer.count");
    
    private void count(String method) {
        counter.info(method);
    }
}
